import wcl = require('../../jswcl-ts');
import * as _ from 'lodash';

export interface INameValue<T> {
    [propName: string]: T
}

export enum RawType {
    string = 'string',
    number = 'number',
    date = 'date',
    time = 'time',
    bool = 'bit',
    image = 'image',
    guid = 'guid'
}

function type1(name: string, rawType?: RawType, options?: IColType): ColType {
    let obj = new ColType(name, rawType, undefined);
    wcl.extends_(obj, options);
    return obj;
}
function type2(name: string, refType: ColType, options?: IColType): ColType {
    let obj = new ColType(name, refType.rawType, refType);
    wcl.extends_(obj, options);
    return obj;
}
interface IColType {
    name?: string;
    length?: number;
    prec?: number;
}
export type IColTypes = INameValue<ColType>;
export function getColeType(name: string): ColType {
    let ret = (ColType as IColumns)[name];
    if (!ret)
        ret = _.find(ColType, i => i instanceof ColType && i.name === name);
    return ret;
}
export class ColType implements IColType {
    name = '';
    rawType = RawType.string;
    refType?: ColType;
    length: number = 0;
    prec: number = 0;
    needQuoted: boolean;
    hasLengthProp: boolean;
    isFixLength: boolean;

    get baseType(): ColType {
        if (this.refType)
            return this.refType.baseType;
        return this;
    }

    constructor(name: string, rawType?: RawType, refType?: ColType) {
        this.name = name;
        this.refType = refType;
        this.isFixLength = false;
        if (refType) {
            this.rawType = refType.rawType;
            this.hasLengthProp = refType.hasLengthProp;
            this.isFixLength = refType.isFixLength;
        }
        else {
            this.rawType = rawType || RawType.string;
            this.hasLengthProp = this.rawType === RawType.string;
            if (name === 'dicimal')
                this.hasLengthProp = true;
            if (name === 'text')
                this.hasLengthProp = false;
            if (name === 'char' || name === 'nchar')
                this.isFixLength = true;
        }
        this.needQuoted = this.rawType !== RawType.number && this.rawType !== RawType.bool;
    }

    toSqlType(length?: number, prec?: number) {
        let baseType = this.baseType;
        length = length || this.length || baseType.length;
        prec = prec || this.prec || baseType.prec;
        if (length > 0 && prec > 0)
            return `${baseType.name}(${length}, ${prec})`;
        else if (length > 0)
            return `${baseType.name}(${length})`;
        else
            return baseType.name;
    }

    //*********************基础类型*********************************/
    static char = type1('char');
    static varchar = type1('varchar');
    static nchar = type1('nchar');
    static nvarchar = type1('nvarchar');
    static text = type1('text');

    static int = type1('int', RawType.number);
    static bigint = type1('bigint', RawType.number);
    static float = type1('decimal', RawType.number);
    static bool = type1('bit', RawType.bool);
    static time = type1('datetime', RawType.time);
    static image = type1('image', RawType.image);
    static guid = type1('uniqueidentifier', RawType.guid);
    //*********************别名类型*********************************/
    static date = type2('date', ColType.char, { length: 10 });
    static shortTime = type2('shortTime', ColType.char, { length: 5 });
    static longId = type2('longId', ColType.bigint);
    static userId = type2('userId', ColType.longId);
    static id = type2('id', ColType.longId);
    static oid = type2('oid', ColType.varchar, { length: 50 });
    static xml = type2('xml', ColType.text);
    static json = type2('json', ColType.text);

    static cdoe1 = type2('c1', ColType.char, { length: 1 });
    static cdoe2 = type2('c2', ColType.char, { length: 2 });
    static cdoe4 = type2('c4', ColType.char, { length: 4 });
    static cdoe8 = type2('c8', ColType.char, { length: 8 });
    static code = type2('code', ColType.char, { length: 8 });

    static float2 = type2('f2', ColType.float, { length: 19, prec: 2 });
    static float4 = type2('f4', ColType.float, { length: 19, prec: 4 });
    static float8 = type2('f8', ColType.float, { length: 19, prec: 8 });

    static text1 = type2('t1', ColType.varchar, { length: 50 });
    static text2 = type2('t2', ColType.varchar, { length: 200 });
    static text3 = type2('t3', ColType.varchar, { length: 500 });
    static text4 = type2('t4', ColType.varchar, { length: 2000 });
    static text5 = type2('t5', ColType.varchar, { length: 8000 });
}

export function col(declare: string, extension?: IColumn): Column {
    return Column.parse(undefined, declare, extension);
}
export let sybOption = Symbol('option');
export let sybIndex = Symbol('index');
export interface IColsOption {
    tableName?: string;
    keys?: IColumn[];
    selectAll?: string;
    logAble?: boolean;
    titles: string[];
}
export function getColsOption(cols: object): IColsOption {
    let ret = cols[sybOption] as IColsOption;
    if (!ret)
        cols[sybOption] = ret = {};
    return ret;
}
export function makeColumns(fromObj: object): IColumns {
    let obj = fromObj as IColumns;
    let option = getColsOption(obj);

    if (obj.hasOwnProperty(0)) {
        obj[sybOption] = obj[0];
        delete obj[0];
    }
    if (wcl.isString(option)) {
        option = { tableName: option };
        obj[sybOption] = option;
    }
    if (option.logAble) {
        obj.createTime = CLogger.createTime;
        obj.updateTime = CLogger.updateTime;
    }

    let keys = Object.keys(obj);
    let pk = [];
    let titles = {};
    let lastGrps: string[] = [];
    for (const key of keys) {
        const column = obj[key];
        if (wcl.isString(column)) {
            obj[key] = col(column as string);
        }
        else if (column instanceof ColType) {
            obj[key] = new Column(key, <ColType>column);
        }
        obj[key].jsName = key;
        if (obj[key].primaryKey)
            pk.push(obj[key]);
        if (!column.fieldName)
            column.fieldName = column.jsName;

        titles[key] = obj[key].caption || obj[key].jsName;

        if (column.index) {
            if (!obj[sybIndex])
                obj[sybIndex] = {};
            obj[sybIndex]['ix_' + column.fieldName] = column.fieldName + (column.unique ? ':uq' : '');
        }

        if (!column.grps)
            column.grps = lastGrps;
        else
            lastGrps = column.grps;
    }
    option.keys = pk;
    option.titles = titles;
    return obj;
}

export class Column {
    jsName?= '';
    fieldName?= '';
    type?: ColType = ColType.text1;
    caption?= '';
    notnull?= false;
    unique?= false;
    index?= false;
    primaryKey?= false;
    defaultValue?: any;
    hint?: string;
    note?: string;
    length?: number;
    prec?: number;
    grps?: string[];

    convert?: (val: any, entity?: any, option?: any) => any;
    check?: (val: any, entity?: any, option?: any) => boolean;

    toSqlType(): string {
        return this.type!.toSqlType(this.length, this.prec);
    }

    constructor(name?: string, type?: ColType) {
        this.fieldName = name || '';
        if (type)
            this.type = type;
    }

    static parse(toColumn?: IColumn, declare?: string, extension?: IColumn): IColumn {
        //varchar(12,3) 'caption' notnull default 21 name='hehe' note '嘻嘻'
        let isNew = !toColumn;
        if (!toColumn)
            toColumn = new Column();
        if (declare) {
            let sr = new wcl.script.reader(declare);

            let curr, name, value;
            if (isNew) {
                curr = sr.popNext();
                if (!curr || !curr.isIdentifyOrString())
                    throw new Error('IColumn.declare 第一节必须是字段类型\r\n' + declare);
                let name = curr.getValue();
                if (name) {
                    toColumn.type = getColeType(name);
                    if (!toColumn.type)
                        throw new Error('IColumn.declare 字段类型非法: ' + name + '\r\n' + declare);
                }
                curr = sr.peekNext();
                if (curr.isArray()) {
                    let value = curr.getValue();
                    if (value.length > 0)
                        toColumn.length = parseInt(value[0]);
                    if (value.length > 1)
                        toColumn.prec = parseInt(value[1]);
                    sr.popNext();

                    curr = sr.peekNext();
                }

                if (curr && !curr.isIdentifyOrString())
                    throw new Error('IColumn.declare 第二节必须是字段标题\r\n' + declare);
                if (curr) {
                    let val = curr.getValue();
                    if (val)
                        toColumn.caption = val;
                    sr.popNext();
                }
            }

            while (true) {
                curr = sr.popNext();
                if (!curr)
                    break;

                if (!curr.isIdentifyOrString())
                    continue;

                name = curr.name;

                if (name === 'notnull')
                    toColumn.notnull = true;
                else if (name === 'pk')
                    toColumn.primaryKey = true;
                else if (name === 'uq')
                    toColumn.unique = true;
                else if (name === 'ix')
                    toColumn.index = true;
                else {
                    value = curr.getValue();
                    if (name === 'field')
                        toColumn.fieldName = value;
                    else if (name === 'default')
                        toColumn.defaultValue = value;
                    else if (name === 'note')
                        toColumn.note = value;
                    else if (name === 'hint')
                        toColumn.hint = value;
                    else if (name === 'grp') {
                        if (value) {
                            value = value.split(',');
                            if (!toColumn.grps)
                                toColumn.grps = value;
                            else
                                toColumn.grps.push(...value);
                        }
                    }
                }
            }
        }
        wcl.extends_(toColumn, extension);
        if (!toColumn.fieldName)
            toColumn.fieldName = toColumn.jsName;

        return toColumn;
    }

    [Symbol.toPrimitive](hint: string) {
        return this.jsName;
    }

    sqlValue(val: any): string | undefined {
        return convertValue(val, this.type!, true);
    }

    valuePart(val: any, withFieldName?: boolean): string {
        val = this.sqlValue(val) as string;
        if (val == undefined)
            val = 'null';
        else if (this.type!.needQuoted) {
            if (wcl.isString(val))
                val = val.replace(replaceQuoted, "''");
            val = val.q;
        }
        return withFieldName ? `[${this.fieldName}] = ${val}` : val;
    }

    nameAndCaption(): string {
        if (!this.caption || this.jsName === this.caption)
            return this.jsName!;
        return `${this.jsName}(${this.caption})`;
    }
}

let replaceQuoted = /'/g;
export function convertToDate(val: any): any {
    if (val == undefined)
        return new Date(0);
    if (wcl.isString(val)) {
        return val.toDate();
    }
    return val;
}

export function convertValue(val: any, toType: ColType, finalAsString?: boolean) {
    if (toType.rawType === RawType.bool)
        return val ? '1' : '0';
    if (toType === ColType.date)
        return Date.prototype.fmt.call(convertToDate(val), 'D') || '';
    if (toType === ColType.time)
        return Date.prototype.fmt.call(convertToDate(val), 'L');

    if (toType.rawType === RawType.string) {
        if (val == undefined)
            return '';
        return String(val);
    }
    if (val != undefined)
        return finalAsString ? String(val) : val;
}

export function getGroupFields(colsObj: IColumns | any, grps: string[]): string[] {
    return _.filter(colsObj, i => i.grps && _.some(grps, j => _.some(i.grps, k => k === j))).map(i => i.jsName!);
}

export let CLogger = {
    createTime: col('datetime 创建时间 grp="-"'),
    updateTime: col('datetime 修改时间'),
}
makeColumns(CLogger);

export interface IColumn extends Column { }
export type IColumns = INameValue<IColumn>;