import wcl = require('../../jswcl-ts');
import * as nodeQ from 'nodeQ';

import * as cols from './column';
import * as dbApi from './db';

export let sybOwner = Symbol('Owner');
export let sybCols = Symbol('Cols');

type IWhere = string | wcl.stringBuilder | object;
type IField = cols.Column | string;
type IFieldFilterFunc = { (field: IField): boolean };
type IFieldFilter = cols.Column | string | (string | cols.Column)[] | IFieldFilterFunc;
type IString = string | wcl.stringBuilder;

function convertDefaultValue(toType: cols.RawType, val: any) {
    if (toType === cols.RawType.number)
        return val || 0;
    if (toType === cols.RawType.bool)
        return val || false;
    if (toType === cols.RawType.string)
        return val || '';

    return val;
}

function isContainField(field: string, filters: IFieldFilter[]): boolean {
    if (field === cols.CLogger.updateTime.jsName)
        return true;
    let bool = true;
    for (let i = 0; i < filters.length; i++) {
        const elem = filters[i];
        if (elem === '-')
            bool = false;
        else if (elem === '+')
            bool = true;
        else {
            if (field === elem)
                return bool;
            if (wcl.isFunction(elem)) {
                if (elem(field))
                    return bool;
            }
            else if (elem instanceof cols.Column) {
                if (field === elem.jsName)
                    return bool;
            }
            else if (wcl.isArray(elem)) {
                if (_.some(elem, i => i === field || (i instanceof cols.Column && i.jsName! === field)))
                    return bool;
            }
        }
    }
    return !bool;
}

let EntityProxyHandle = {
    getOwner(target: object): Entity {
        return target[sybOwner];
    },

    get(target: IEntity, property: PropertyKey) {
        let theCol: cols.Column;
        if (property instanceof cols.Column) {
            theCol = property;
            property = property.jsName;
        }
        else if (wcl.isString(property)) {
            let owner = this.getOwner(target);
            if (property === 'hasOwnProperty')
                return Object.prototype.hasOwnProperty.bind(owner.raw);
            theCol = owner.entityType.cols[property];

            if (!theCol) {
                if (property[0] === '$')
                    property = property.substr(1);
                if (property === 'ow')
                    return owner;

                let val = owner[property] as Function;
                if (val && wcl.isFunction(val))
                    return val.bind(owner);

                if (property === 'toJSON' || property === 'then') //调用JSON.stringify会查询toJSON值，返回空从而使用默认的JSON序列化行为。then是Promise对象被访问的值。
                    return;

                if (!owner.hasOwnProperty(property) && !owner.__proto__.hasOwnProperty(property))
                    throw new Error(`表'${owner.entityType.tableName}'不存在字段: '${property}(区分大小写)'`)

                return val;
            }
        }

        let val = Reflect.get(target, property);
        if (!theCol || !(theCol instanceof cols.Column))
            return val;
        val = convertDefaultValue(theCol.type!.rawType, val);
        if (theCol.type!.isFixLength && wcl.isString(val))
            val = val.trimRight();

        return val;
    },

    set(target: object, property: PropertyKey, val: any): boolean {
        let theCol: cols.Column;
        let owner = this.getOwner(target);
        if (property instanceof cols.Column) {
            theCol = property;
            property = property.jsName;
        }
        else if (wcl.isString(property)) {
            theCol = owner.entityType.cols[property];

            if (!theCol) {
                if (property[0] === '$')
                    property = property.substr(1);

                if (!owner.hasOwnProperty(property) && !owner.__proto__.hasOwnProperty(property))
                    throw new Error(`表'${owner.entityType.tableName}'不存在字段: '${property}(区分大小写)'`)

                Reflect.set(owner, property, val);
                return true;
            }
        }

        if (owner.state === EntityState.normal)
            owner.edit();

        if (owner.state === EntityState.edit && !owner.oldValues.hasOwnProperty(property))
            owner.oldValues[property] = Reflect.get(target, property);

        val = convertDefaultValue(theCol.type!.rawType, val);
        return Reflect.set(target, property, val);
    },

    // has(target: object, property: PropertyKey) {
    //     return Reflect.has(target, property);
    // }
}

export enum EntityState {
    new = '',
    edit = '',
    delete = '',
    normal = '',
}
wcl.defConsts(EntityState);

function hi(db?: nodeQ.MSSql): nodeQ.MSSql {
    return db || dbApi.main;
}

export interface IExecuteOption {
    db?: dbApi.IDb;
    trans?: dbApi.Transaction;
}
export interface IUpdateOption extends IExecuteOption {
    onlyModified?: boolean;
    fields?: IFieldFilter[];
}

interface IStaticEntity {
    readonly cols: cols.IColumns;
    readonly tableName: string;
    readonly primaryKeys: cols.IColumn[];
}
class StaticEntity {
    static get cols(): cols.IColumns {
        return this[sybCols] as cols.IColumns;
    }

    static get titleCols(): cols.INameValue<string> {
        return this[sybCols][cols.sybOption].titles;
    }

    static get tableName(): string {
        let theCols = this.cols;
        let option = cols.getColsOption(theCols);
        return option.tableName!;
    }

    static get primaryKeys(): cols.IColumn[] {
        let theCols = this.cols;
        let option = cols.getColsOption(theCols);
        return option.keys!;
    }

    static async getSelect(where: IWhere) {
        return getSelect(this, where);
    }

    static async select(where: IWhere, option: IExecuteOption = wcl.empty): Promise<nodeQ.DbRecordSet> {
        let sql = getSelect(this, where);
        return hi(option.db).qry(sql);
    }

    static async one<T>(where: IWhere, option: IExecuteOption = wcl.empty): Promise<T> {
        let r = await this.select(where, option);
        return r && r.length > 0 ? this.edit(r[0]) : undefined;
    }

    static async id<T>(value: any, option: IExecuteOption = wcl.empty): Promise<T> {
        let theCols = this.cols;
        let colsOption = cols.getColsOption(theCols);
        let where = {};
        where[colsOption.keys[0]] = value;
        return await this.one<T>(where, option);
    }

    static async key<T>(colOrColumnName: string | cols.Column, value: any, option: IExecuteOption = wcl.empty): Promise<T> {
        let theCols = this.cols;
        let colsOption = cols.getColsOption(theCols);
        let where = {};
        where[colOrColumnName] = value;
        return await this.one<T>(where, option);
    }

    static async all<T>(where: IWhere, option: IExecuteOption = wcl.empty): Promise<T[]> {
        let rs: object[] = await this.select(where, option);
        let ret = rs && rs.map(i => this.edit(i), this);
        ret && (ret.columns = this.titleCols);
        return ret;
    }

    static getInsert(insertContent: object): string {
        let obj: Entity = new this(insertContent, true);
        return obj.getInsert();
    }
    static async insert(insertContent: object, option: IExecuteOption = wcl.empty): Promise<number> {
        let sql = this.getInsert(insertContent);
        return hi(option.db).exe(sql, option.trans);
    }

    static getUpdate(updateContent: object, where?: IWhere, option: IUpdateOption = wcl.empty): string {
        // where = getWhere(this, where, true);
        // if (!where)
        //     throw new Error('因为update是个危险的动作，必须添加where条件语句，请检查你的条件语句!');
        let sql = getUpdate(this, updateContent, where, option.fields);
        return sql;
    }
    static async update(updateContent: object, where?: IWhere, option: IUpdateOption = wcl.empty): Promise<number> {
        let sql = this.getUpdate(updateContent, where, option);
        return hi(option.db).exe(sql, option.trans);
    }

    static getDelete(where: IWhere): string {
        where = getWhere(this, where);
        if (!where)
            throw new Error('因为delete是个危险的动作，必须添加where条件语句，请检查你的条件语句!');
        return `delete [${this.tableName}] where ${where}`;
    }
    static async delete(where: IWhere, option: IExecuteOption = wcl.empty): Promise<number> {
        let sql = this.getDelete(where);
        return hi(option.db).exe(sql, option.trans);
    }

    static new(rawObj?: object): object {
        return new this(rawObj, true).val;
    }

    static edit(rawObj: object): object {
        return new this(rawObj, false).val;
    }

    static trimFixLength<T>(rawOrEntity: T): T {
        if (!rawOrEntity)
            return;
        let cols = this.cols;
        for (const key in rawOrEntity) {
            // if (Object.prototype.hasOwnProperty.call(rawOrEntity, key)) {
            if (rawOrEntity.hasOwnProperty(key)) {
                if (cols[key] && cols[key].type!.isFixLength) {
                    const elem = rawOrEntity[key];
                    if (wcl.isString(elem))
                        rawOrEntity[key] = elem.trimRight();
                }
            }
        }
        return rawOrEntity;
    }
}

export class Entity extends StaticEntity {
    /**实际存储实体数据的空间对象*/
    raw?: object;
    /**缓存本次修前的各属性值（在提交后会清空此缓存） */
    oldValues?: object;
    /**实体存储空间，raw对象的代理。 */
    val?: Object;
    /**实体类的修改状态 */
    state: EntityState = EntityState.normal;
    readonly ow!: Entity;//仅仅是一个给智能提示用的占位符，实际取值在Proxy对象里已处理。

    constructor(rawObj?: object, isNew?: boolean) {
        super();
        this.touch(rawObj, isNew);
    }

    get entityType(): IStaticEntity {
        return this.__proto__.constructor;
    }

    touch(rawObj?: object, isNew?: boolean): object {
        if (isNew && rawObj) {
            this.raw = wcl.extends({}, rawObj);
        }
        else if (isNew)
            this.raw = {};
        else if (rawObj)
            this.raw = rawObj;
        else
            throw new Error(`准备使用touch或者edit修改对象时，原始对象rawObj不能为空!`);
        this.raw[sybOwner] = this;
        this.oldValues = undefined;
        this.val = new Proxy(this.raw, EntityProxyHandle);
        this.state = isNew || !rawObj ? EntityState.new : EntityState.normal;

        return this.val;
    }

    drop(): void {
        if (this.state !== EntityState.normal)
            throw `状态为'${this.state}'的对象，不允许调用drop方法`;
        this.state = EntityState.delete;
    }

    edit(): void {
        if (this.state !== EntityState.normal)
            throw `状态为'${this.state}'的对象，不允许调用edit方法`;
        this.state = EntityState.edit;
        this.oldValues = {};
    }

    get isChanged(): boolean {
        return this.state === EntityState.edit &&
            Boolean(this.oldValues) && Object.keys(this.oldValues).length > 0;
    }

    isModified(col: IField): boolean {
        if (!this.oldValues)
            return false;
        if (col instanceof cols.Column)
            col = col.jsName;
        return this.oldValues.hasOwnProperty(col);
    }

    accept(): void {
        this.state = EntityState.normal;
        this.oldValues = undefined;
    }

    reject(property: PropertyKey | cols.Column): void {
        if (this.state !== EntityState.edit)
            return;

        if (property) {
            if (property instanceof cols.Column)
                property = property.jsName!;
            if (this.oldValues!.hasOwnProperty(property)) {
                Reflect.set(this.raw!, property, this.oldValues[property]);
                delete this.oldValues[property];
            }
        }
        else {
            wcl.extends_(this.raw, this.oldValues);
            this.state = EntityState.normal;
            this.oldValues = undefined;
        }
    }

    getPrimaryKeyPart(): string {
        return getPrimaryKeyPart(this.entityType.primaryKeys, this.raw, true);
    }

    getInsert(): string {
        let theCols = this.entityType.cols;
        let option: cols.IColsOption = theCols[cols.sybOption];
        if (option.logAble)
            this.val[cols.CLogger.createTime.jsName] = new Date();
        if (option.keys[0].type!.baseType === cols.ColType.bigint
            && !this.val[option.keys[0]])
            this.val[option.keys[0]] = wcl.newId;

        let sb = new wcl.stringBuilder();
        sb.append('insert into [', this.entityType.tableName, '] (');
        for (const key in theCols) {
            let elem: cols.Column = theCols[key];
            sb.append('[', elem.fieldName, '], ');
        }
        sb.backChar(2);
        sb.append(') values (');
        let col: cols.Column;
        try {
            for (const key in theCols) {
                col = theCols[key];
                sb.append(col.valuePart(this.val[col.jsName], false), ', ');
            }
        }
        catch (err) {
            throw new Error(`获取字段'${col.jsName}'的值时出现错误\r\n${err.message}`);
        }
        sb.back();
        sb.append(')');
        return sb.toString();
    }

    getUpdate(option: IUpdateOption = wcl.empty): string {
        let onlyModified = option.onlyModified;
        if (onlyModified == undefined)
            onlyModified = true;
        let theCols = this.entityType.cols;

        let colOption: cols.IColsOption = theCols[cols.sybOption];
        if (colOption.logAble)
            this.val[cols.CLogger.updateTime.jsName] = new Date();

        let fields = toCols(theCols, option.fields);

        let sb = new wcl.stringBuilder();
        sb.append('update [', this.entityType.tableName, '] set ');
        let found = false;

        let col: cols.Column;
        try {
            for (const key in theCols) {
                col = theCols[key];
                if (col.primaryKey)
                    continue;
                if (onlyModified && !this.isModified(col))
                    continue;
                if (fields && !isContainField(col.jsName!, fields))
                    continue;
                found = true;
                sb.append(col.valuePart(this.val[col.jsName], true), ', ');
            }
        }
        catch (err) {
            throw new Error(`获取字段'${col.jsName}'的值时出现错误\r\n${err.message}`);
        }
        if (!found)
            return '';
        sb.back();
        sb.append(' where ', this.getPrimaryKeyPart());
        return sb.toString();
    }

    getDelete(): string {
        return `delete [${this.entityType.tableName}] where ${this.getPrimaryKeyPart()}`;
    }

    getActionSql(): string {
        switch (this.state) {
            case EntityState.new: return this.getInsert();
            case EntityState.edit: return this.getUpdate();
            case EntityState.delete: return this.getDelete();
        }
    }

    async insert(option: IExecuteOption = wcl.empty): Promise<number> {
        let sql = this.getInsert();
        let ret = await hi(option.db).exe(sql, option.trans);
        this.accept();
        return ret;
    }

    async update(option: IUpdateOption = wcl.empty): Promise<number> {
        let sql = this.getUpdate(option);
        let ret = await hi(option.db).exe(sql, option.trans);
        this.accept();
        return ret;
    }

    async delete(option: IExecuteOption = wcl.empty): Promise<number> {
        let sql = this.getDelete();
        let ret = await hi(option.db).exe(sql, option.trans);
        return ret;
    }

    async commit(option: IExecuteOption = wcl.empty): Promise<number> {
        let sql = this.getActionSql();
        let ret = await hi(option.db).exe(sql, option.trans);
        this.accept();
        return ret;
    }
}
// let hide: PropertyDescriptor = { enumerable: false, writable: true };
// Object.defineProperty(Entity.prototype, 'raw', hide);
// Object.defineProperty(Entity.prototype, 'state', {});
// Object.defineProperty(Entity.prototype, 'oldValues', hide);

export interface EntityClass extends IStaticEntity { }
type EntityObj = EntityClass | cols.IColumns | object;

function getPrimaryKeyPart(keys: cols.IColumn[], content: object, mustExists: boolean): string {
    let r = '';
    for (const col of keys) {
        if (mustExists && !content.hasOwnProperty(col.jsName!))
            throw new Error(`条件语句中未指定主键字段'${col.jsName}'的值!`);
        if (r)
            r += ' and ';
        r += col.valuePart(content[col.jsName!], true);
    }
    return r;
}

function getWhere(entityClass: EntityObj, where?: IWhere): string {
    let theCols = entityClass[sybCols] as cols.IColumns || entityClass;
    if (wcl.isObject(where) && !(where instanceof wcl.stringBuilder)) {
        let sb = wcl.sb();
        for (const key in where) {
            let theCol = theCols[key];
            if (!theCol)
                continue;
            sb.append('[', theCol.fieldName!, '] = ', theCol.valuePart(where[key]), ' and ');
        }
        sb.back();
        where = sb;
    }
    return where && where.toString();
}

function getSelect(entityClass: EntityObj, where?: IWhere): string {
    let theCols = entityClass[sybCols] as cols.IColumns || entityClass;
    let option = cols.getColsOption(theCols);
    let all = option.selectAll;
    if (!all) {
        let sb = new wcl.stringBuilder();
        sb.append('select ');
        for (const key in theCols) {
            const elem = theCols[key];
            if (elem.fieldName === elem.jsName)
                sb.append('[', elem.fieldName!, '], ');
            else
                sb.append('[', elem.fieldName!, '] as ', elem.jsName!, ', ');
        }
        sb.backChar(2);
        sb.append(' from [', option.tableName!, ']');

        all = option.selectAll = sb.toString();
    }
    where = getWhere(entityClass, where);
    if (!where)
        return all;
    return all + ' where ' + where;
}

function getUpdate(entityClass: EntityObj, updateContent: object, where?: IWhere, fields?: IFieldFilter[]): string {
    let theCols = entityClass[sybCols] as cols.IColumns || entityClass;
    let option = cols.getColsOption(theCols);
    if (option.logAble)
        updateContent[cols.CLogger.updateTime.jsName] = new Date();
    fields = toCols(theCols, fields);

    let sb = new wcl.stringBuilder();
    sb.append('update [', option.tableName!, '] set ');
    let found = false;

    let col: cols.Column;
    try {
        for (const key in updateContent) {
            col = theCols[key];            
            if (!col || col.primaryKey)
                continue;
            if (fields && !isContainField(col.jsName!, fields))
                continue;
            found = true;
            sb.append(col.valuePart(updateContent[key], true), ', ');
        }
    }
    catch (err) {
        throw new Error(`获取字段'${col.jsName}'的值时出现错误\r\n${err.message}`);
    }
    if (!found)
        return '';
    sb.back();
    sb.append(' where ');
    if (where && (where = getWhere(entityClass, where)))
        sb.append(where);
    else
        sb.append(getPrimaryKeyPart(option.keys!, updateContent, true));

    return sb.toString();
}

function toCols(theCols: cols.IColumns, fields: IField[]): cols.Column[] {
    if (!fields)
        return;
    for (let i = 0; i < fields.length; i++) {
        const elem = fields[i];
        if (wcl.isString(elem) && elem !== '-' && elem !== '+')
            fields[i] = theCols[elem];
    }
    return fields;
}

function toFields(fields: IField[]): string[] {
    if (!fields)
        return;
    for (let i = 0; i < fields.length; i++) {
        const elem = fields[i];
        if (elem instanceof cols.Column)
            fields[i] = elem.jsName!;
    }
    return fields;
}