import wcl = require('../../jswcl-ts');
import _ = require('lodash');

import * as nodeQ from 'nodeQ';
import * as cols from './column';
import * as dbApi from './db';
import * as ent from './entity';

interface IFieldnameResult {
    TableName: string,
    FieldName: string,
    FieldPos: number,
    UserType: string,
    Type: string,
    length: number,
    xprec: number,
    xscale: number,
    isnullable: boolean,
}
function getAllFields(tableName: string): string {
    return `SELECT
            t.name AS 'TableName',
            c.name AS 'FieldName',
            c.colid As 'FieldPos',
            p.name AS UserType,
            p2.name AS Type,
            c.length,
            c.xprec,
            c.xscale,
            c.isnullable
        FROM
            sysobjects t
            JOIN Syscolumns c ON t.id = c.id
            JOIN systypes p ON c.xusertype = p.xusertype
            JOIN systypes p2 ON p.xtype = p2.xtype AND p.xtype = p2.xusertype
        WHERE
            t.xtype = 'U' AND t.name ='${tableName}'
        ORDER BY
            t.name, c.name`;
}

// let nologOption = { nolog: 1 };
function nolog(sql: string): wcl.cfgableString {
    return new wcl.cfgableString(sql);
}


export function getCreateTable(entityClass: ent.EntityClass,
    option: { dropFirst?: boolean, ifNotExists?: boolean } = wcl.empty): string[] {

    let theCols = entityClass[ent.sybCols] as cols.IColumns || entityClass;
    let colOption = cols.getColsOption(theCols);

    let sb = wcl.sb();
    if (option.dropFirst) {
        sb.appendLine(`IF OBJECT_ID ('dbo.${colOption.tableName}') IS NOT NULL
    DROP TABLE dbo.[${colOption.tableName}];`);
    }
    else if (option.ifNotExists) {
        sb.appendLine(`IF OBJECT_ID ('dbo.${colOption.tableName}') IS NULL`);
    }
    sb.appendLine('create table [', colOption.tableName!, '] (');
    for (const key in theCols) {
        const col = theCols[key];
        sb.append(`   [${col.fieldName}] \t\t ${col.toSqlType()}`);
        if (col.notnull)
            sb.append(' not null,\r\n');
        else
            sb.append(',\r\n');
    }
    sb.append(`   constraint PK_${colOption.tableName!} primary key (`);
    for (const key in colOption.keys!) {
        const col = colOption.keys![key];
        sb.append(col.fieldName!, ', ');
    }
    sb.back();
    sb.appendLine(')');
    sb.appendLine(')');

    let r = [];
    r.push(sb.toString());
    sb.clear();

    let indexes = theCols[cols.sybIndex];
    if (indexes) {
        for (const key in indexes) {
            let fields = indexes[key].left(':');
            let opt = indexes[key].right(':').trim(true, false);
            let where;
            if (opt !== 'uq')
                where = opt;
            sb.appendLine(`CREATE ${opt ? 'UNIQUE' : ''} INDEX ${key} ON dbo.[${colOption.tableName}] (${fields}) ${where};`);
        }
    }

    r.push(sb.toString());

    return r;
}


export interface IUpgradeDbOption {
    forceRecreate?: boolean;
    isWush?: boolean;
    autoWithStructDb?: boolean;
}
export async function upgrade(entityClass: ent.EntityClass, db: nodeQ.MSSql, option: IUpgradeDbOption = {}) {
    let theCols = entityClass[ent.sybCols] as cols.IColumns || entityClass;
    let colOption = cols.getColsOption(theCols);
    let sql: string;

    if (option.forceRecreate) {
        let sqls = getCreateTable(theCols, { dropFirst: true });
        for (const i of sqls) {
            await db.exe(nolog(i));
        }
        return 1;
    }
    else {
        sql = `SELECT CASE WHEN OBJECT_ID ('dbo.${colOption.tableName}') > 0 THEN 1 ELSE 0 END`;
        let exists = await db.val(nolog(sql));
        let count = 0;
        if (exists)
            count = await db.val(nolog(`select count(1) from [${colOption.tableName}]`));
        if (!exists || count === 0) {
            let sqls = getCreateTable(theCols, { dropFirst: exists });
            for (const i of sqls) {
                await db.exe(nolog(i));
            }
            return 1; //create table
        }
    }

    sql = getAllFields(colOption.tableName!);
    let rs = await db.qry(nolog(sql)) as nodeQ.IDbRecordSet<IFieldnameResult>;

    let updateFields = '';
    for (const key in theCols) {
        const elem = theCols[key];
        let field = _.find(rs, i => i.FieldName === elem.fieldName);
        if (!field) {
            let notNull = '';
            if (elem.notnull)
                notNull = elem.type!.needQuoted ? "not null default ''" : "not null default 0";
            let add = `alter table [${colOption.tableName}] add [${elem.fieldName}] ${elem.toSqlType()} ${notNull}`;
            updateFields += add + '\r\n';

            continue;
        }

        let newType = elem.type!.baseType;
        if (newType.name !== field.Type) {
            log(`无法自动升级表'${colOption.tableName}'的字段'${elem.fieldName}'类型: ${field.Type}->${newType.name}`.warn);
            continue;
        }

        newType = elem.type!;
        if (newType.hasLengthProp) {
            let isFloat = newType.prec !== 0 || field.xscale !== 0;
            if (isFloat ? newType.length !== field.xprec || newType.prec !== field.xscale : newType.length !== 0 && newType.length !== field.length) {
                let notNull = '';
                if (elem.notnull)
                    notNull = elem.type!.needQuoted ? "not null default ''" : "not null default 0";
                let alter = `alter table [${colOption.tableName}] alter column [${elem.fieldName}] ${elem.toSqlType()} ${notNull}`;
                updateFields += alter + '\r\n';
                continue;
            }
        }
    }
    _.each(rs, field => {
        if (!_.some(theCols, i => i.fieldName === field.FieldName))
            log(`无法自动升级表'${colOption.tableName}'，删除字段'${field.FieldName}'`.warn);
    });
    if (updateFields) {
        log('auto update fields...');
        log(updateFields);
        await db.exe(nolog(updateFields));
    }
}

let classList: cols.INameValue<ent.EntityClass> = {};
export function reg(entityClass: ent.EntityClass): void {
    classList[entityClass.tableName] = entityClass;
}
export function get(entityClassName: string, errIfNotfount: boolean = true): ent.EntityClass {
    let ret = classList[entityClassName];
    if (!ret && errIfNotfount)
        throw new Error(`没有发现已注册的实体类型名称: ${entityClassName}`);
    return ret;
}

export async function upgradeAll(db?: nodeQ.MSSql, option: IUpgradeDbOption = {}): Promise<void> {
    if (dbApi.main.hostId === dbApi.struct.hostId && option.forceRecreate && !option.isWush)
        throw new Error(`为了防止误操作, 不允许在主库和结构库是同一个库(${dbApi.main.hostId})的时候执行强制重建的工作!`);

    db = db || dbApi.struct;
    for (const i in classList) {
        await upgrade(classList[i], db, option);
    }
}

export async function autoUpgradeAll(db?: nodeQ.MSSql, option: IUpgradeDbOption = {}): Promise<void> {
    if (db && !db.hostId) {
        option = db;
        db = undefined;
    }

    if (option.autoWithStructDb) {
        let tmpOption: IUpgradeDbOption = wcl.extends({}, option);
        tmpOption.forceRecreate = true;
        await upgradeAll(dbApi.struct, tmpOption);
    }

    if (!dbApi.isAutoUpgrade)
        return;
    db = db || dbApi.main;
    if (db !== dbApi.struct || !option.autoWithStructDb)
        await upgradeAll(db, option);
}

export * from './column';
export * from './entity';
export { dbApi as db };

interface IMakeOption {
    reg?: boolean;
    logAble?: boolean;
}
export function make(tableName: string, theCols: cols.IColumns | any, entityType: ent.EntityClass, option: IMakeOption = wcl.empty) {
    let colOption: cols.IColsOption = theCols[cols.sybOption] || {};
    colOption.logAble = option.logAble;
    colOption.tableName = tableName;
    theCols[cols.sybOption] = colOption;
    cols.makeColumns(theCols);

    entityType[ent.sybCols] = theCols;
    if (option.reg == undefined || option.reg)
        exports.reg(entityType);

    _.each(entityType, (i, key) => {
        if (wcl.isArray(i) && i[0] === '@grp') {
            i.shift();
            entityType[key] = cols.getGroupFields(theCols, i);
        }
    });
}