const char = require('./char');
const wcl = require('./wcl');

let empty = {};
let tokenType = exports.tokenType = {
    string: '',
    int: '',
    float: '',
    identify: '',
    remark: '',
    symbol: '',
    whiteSpace: '',
    array: '',
    keyword: '',
}
wcl.defConsts(tokenType);

let token = exports.token = class token {
    constructor(theTokenType, value, name) {
        this.type = theTokenType;
        this.value = value;
        this.name = name;
        if (theTokenType !== tokenType.keyword)
            this.hasValue = theTokenType == tokenType.identify ? false : value && true;
    }

    isSymbol(value) {
        return this.type === tokenType.symbol && (!value || this.value === value);
    }

    isIdentify(name) {
        return this.type === tokenType.identify && (!name || this.name === name);
    }

    isIdentifyOrString(name) {
        if (this.isIdentify(name))
            return true;
        return this.type === tokenType.string && (name === undefined || name === this.value);
    }

    isArray() {
        return this.type == tokenType.array;
    }

    isKeyword() {
        return this.type === tokenType.keyword;
    }

    getValue() {
        if (!this.hasValue)
            return this.name;
        let value = this.value;
        if (value && value instanceof token)
            return value.getValue();
        if (this.type === tokenType.array && value) {
            let rvalue = new Array(value.length);
            for (let i = 0; i < rvalue.length; i++) {
                rvalue[i] = value[i].getValue();
            }
            return rvalue;
        }
        return value;
    }

    get isValue() {
        let type = this.type;
        return type === tokenType.string || type === tokenType.int || type === tokenType.float
            || type === tokenType.identify || type === tokenType.array || (type === tokenType.keyword && this.hasValue);
    }

    toString() {
        let value = this.type === tokenType.array ? ''.reset + '(' + this.value.join(', ') + ')'.reset :
            '"' + wcl.valueInfo(this.value) + '"';

        let name = this.name ? ' <' + this.name + '> = ' : '';
        let ret = `${''.reset}${this.type}:${name.blue} ${value.data}`;
        return ret;
    }
}

class readerCore {
    /**
     * 
     * @param {string} src 
     */
    constructor(src, start = 0) {
        this.src = src || '';
        this.pos = this.start = start || 0;
        this.arrDeep = 0;
    }

    isEnd(offset = 0) {
        offset += this.pos;
        return offset < 0 || offset >= this.src.length;
    }
    get curr() {
        return this.next();
    }
    next(offset = 0) {
        return this.isEnd(offset) ? undefined : this.src[this.pos + offset];
    }
    value(start, end) {
        return this.src.substr(start, this.pos - start + (end || 0));
    }
    get explainTo() {
        return this.src.substr(this.start, this.pos);
    }

    readString(startChar) {
        let start = this.pos++;
        let saveIndex = start + 1;
        let enableEscape = startChar === '`';

        //多数字符串不需要转义，所以这里先优化处理（先判断不需要转义的字符串）
        while (!this.isEnd()) {
            let ch = this.curr;
            if (ch === startChar) {//整段字符段不需要转义，则直接拷贝整段。
                this.pos++;
                return this.value(start + 1, -1);
            }
            if (enableEscape && ch === '\\') //在第一次出现转义字符进入转义处理。
                break;
            this.pos++;
        }

        if (enableEscape) {
            let sb = new wcl.stringBuilder();
            while (!this.isEnd()) {
                let ch = this.curr;
                if (ch === startChar) {
                    sb.append(this.value(saveIndex));
                    this.pos++;
                    return sb.toString();
                }
                if (ch === '\\') {
                    sb.append(this.value(saveIndex));
                    this.pos++;
                    ch = this.curr;
                    if (ch === 'r')
                        sb.append('\r');
                    else if (ch === 'n')
                        sb.append('\n');
                    else if (ch === 't')
                        sb.append('\t');
                    else
                        sb.append(ch);
                    saveIndex = this.pos + 1;
                }
                this.pos++;
            }
        }

        throw new Error('字符串没有结束！\r\n' + this.value(start));
    }

    readNumber() {
        let dot = false;
        let start = this.pos;
        this.pos++;
        while (!this.isEnd()) {
            let ch = this.curr;
            if (ch === '.') {
                if (dot)
                    throw new Error('错误的数值: "' + this.value(start, 1) + '"');
                dot = true;
            }
            else if (!char.digit.test(ch))
                break;
            this.pos++;
        }
        let text = this.value(start);
        return [dot ? parseFloat(text) : parseInt(text), dot];
    }

    readIdentify() {
        let start = this.pos;
        this.pos++;
        while (!this.isEnd() && char.letterOrDigit.test(this.curr)) {
            this.pos++;
        }
        let text = this.value(start);
        if (text === 'null')
            return [true, text, undefined, true];
        else if (text === 'false')
            return [true, text, false, true];
        else if (text === 'true')
            return [true, text, true, true];
        else if (text === 'now')
            return [true, text, new Date(), true];
        return [false, text];
    }

    readSymbol(startChar) {
        let next = this.next(1);
        if (startChar === '/' && next === '/') {
            this.pos += 2;
            return '//';
        }
        else if (startChar === '/' && next === '*') {
            this.pos += 2;
            return '/*';
        }
        this.pos++;
        return startChar;
    }

    readArray() {
        let r = [];
        let start = this.pos - 1;
        this.arrDeep++;
        while (!this.isEnd()) {
            let next = this.readNext();
            if (!next)
                break;
            if (next.isSymbol(')'))
                return r;
            r[r.length] = next;
        }
        throw "数组没有结束\r\n" + this.value(start);
    }

    readRemark(startChar) {
        // let start = this.pos;
        this.pos += startChar.length;
        if (startChar === '//') {
            let idx = this.src.indexOf('\n', this.pos);
            if (idx > 0)
                this.pos = idx + 1;
        }
        else if (startChar === '/*') {
            let idx = this.src.indexOf('*/', this.pos);
            if (idx > 0)
                this.pos = idx + 2;
        }
    }

    readWhitespace() {
        // let start = this.pos;
        this.pos++;
        while (!this.isEnd() && char.whiteSpace.test(this.curr)) {
            this.pos++;
        }
    }

    /**@type {token} */
    readNext() {
        if (this.isEnd())
            return;
        let ch = this.src[this.pos];

        if (char.whiteSpace.test(ch)) {
            this.readWhitespace();
            return this.readNext();
        }
        else if (char.letter.test(ch)) {
            let [isKeyword, name, value, hasValue] = this.readIdentify();
            let newToken = new token(isKeyword ? tokenType.keyword : tokenType.identify,
                value, name);
            if (hasValue)
                newToken.hasValue = true;
            return newToken;
        }
        else if (ch === '\'' || ch === '"' || ch === '`') {
            let value = this.readString(ch);
            return new token(tokenType.string, value);
        }
        let nextChar = this.curr;
        if (char.digit.test(ch) || (nextChar && ch === '-' && char.digit.test(nextChar))) {
            let [value, isFloat] = this.readNumber();
            if (isFloat)
                return new token(tokenType.float, value);
            else
                return new token(tokenType.int, value);
        }
        else if (char.symbol.test(ch)) {
            let value = this.readSymbol(ch);
            if (value === ',') {
                return this.readNext();
            }
            if (value === '#') {
                let a = new token(tokenType.identify, undefined, undefined);
                a.hasValue = true;
                return a;
            }
            if (value === '(') {
                value = this.readArray();
                return new token(tokenType.array, value);
            }
            if (value === ')') {
                if (this.arrDeep)
                    this.arrDeep--;
                else
                    throw new Error('错误的数组结束符: \r\n' + this.value(this.start));
            }
            if (value === '//' || value == '/*') {
                this.readRemark(value);
                return this.readNext();
            }
            return new token(tokenType.symbol, value);
        }
    }
}

exports.reader = class scriptReader {
    constructor(src, start = 0) {
        this.__core = new readerCore(src, start);
        /**@type {token} */
        this.__lastRead = undefined;

        /**@type {token[]} */
        this.__arr = [];

        this.nameParameterSymbol = '=';
    }

    /**@returns {token} */
    readNext() {
        /**@type {token} */
        let curr = this.__lastRead || this.__core.readNext();
        this.__lastRead = undefined;

        if (curr && curr.isIdentify()) {
            let next = this.__core.readNext();
            if (next && next.isSymbol(this.nameParameterSymbol)) {
                let value = this.readNext();
                if (!value || !value.isValue)
                    throw new Error('此处需要一个值\r\n' + curr.name + '=');
                curr.value = value;
                curr.hasValue = true;
            }
            else
                this.__lastRead = next;
        }
        return curr;
    }


    /**@returns {token} */
    popNext(count = 1) {
        for (let i = 1; i < count; i++) {
            this.__arr.shift();
        }
        if (this.__arr.length > 0)
            return this.__arr.shift();
        return this.readNext();
    }

    /**@returns {token} */
    peekNext(pos = 0) {
        for (let i = this.__arr.length; i <= pos; i++) {
            let next = this.readNext();
            if (next)
                this.__arr.push(next);
            else
                return
        }
        if (this.__arr.length > pos)
            return this.__arr[pos];
    }
}

exports.test = function () {
    let text = `value='wush"' 12, 34.45, (21, 'wush', 34)hehe() //wush is ok`;
    text = `name=null,(null, wush, 'hehe', (21, now, name, true, 'xixi', false, null)) nick=name, age=32 addres='yintan' (((32, 545, 'wush'))) varchar(20, 4) notnull caption='haha is ok!' default=22 pk autoId`;
    let reader = new wcl.script.reader(text);
    // reader.nameParameterSymbol = ':';
    while (true) {
        let curr = reader.popNext();
        if (!curr)
            break;
        // log('value: '.red , curr.getValue());
        if (curr.isIdentify() && !curr.hasValue) {
            let next = reader.peekNext();
            if (next && next.isArray()) {
                log(''.red, `function: ${curr.name}(${next.value})`);
                reader.popNext();
                continue;
            }
        }
        log(curr.toString());
    }
}